import numpy as np
from PIL import Image

def read_idx_file(fileName):
	with open(fileName, "rb") as file:
		magic = np.fromfile(file, dtype=np.uint8, count=4)
		if (magic[0]!=0 or magic[1]!=0):
			return None
		dtype = magic[2]
		if (dtype==0x08): dtype=np.uint8
		else: 
			if (dtype==0x09): dtype=np.int8
			else: 
				if (dtype==0x0B): dtype=np.int16
				else: 
					if (dtype==0x0C): dtype=np.int32
					else: 
						if (dtype==0x0D): dtype=np.float
						else: 
							if (dtype==0x0E): dtype=np.double
		# print(dtype)
		dim_count = int(magic[3])
		# print(dim_count)
		dims = []
		bytes = np.fromfile(file, dtype='>i4', count=dim_count)
		dims = list(bytes)
		# print(dims)

		item_count=1
		for x in dims:
			item_count *= x

		# print(item_count)
		items = np.fromfile(file, dtype=dtype, count=item_count)
		
		data = np.reshape(items, tuple(dims), order='C')

		# print(data.shape)
		# print(data)
		return data


def make_image_matrix(rows, cols, images, fileName):
	img_height = images.shape[1]
	img_width = images.shape[2]
	# print(img_width,  img_height)
	mat_width = cols*img_width
	mat_height = rows*img_height
	matrix = Image.new('L', (mat_width, mat_height))
	for row in range(rows):
		for col in range(cols):
			idx = row*cols + col
			data = images[idx].astype(np.int32)
			img = Image.fromarray(data, mode='I')
			matrix.paste(img, box=(col*img_width, row*img_height))
	matrix.save(fileName)

# result (id, results)
def make_image_result(results, query_images, dataset_images, fileName):
	img_height = query_images.shape[1]
	img_width = query_images.shape[2]
	cols = len(results[0][1])+2
	rows = len(results)
	# print(img_width,  img_height)
	mat_width = cols*img_width
	mat_height = rows*img_height
	matrix = Image.new('L', (mat_width, mat_height))
	
	for row, result in enumerate(results):
		# query image
		data = query_images[result[0]].astype(np.int32)
		img = Image.fromarray(data, mode='I')
		matrix.paste(img, box=(0, row*img_height))
		for col in range(2, cols, 1):
			idx = result[1][col-2]
			data = dataset_images[idx].astype(np.int32)
			img = Image.fromarray(data, mode='I')
			matrix.paste(img, box=(col*img_width, row*img_height))
	
	matrix.save(fileName)

indices10 = read_idx_file("t10k-labels.idx1-ubyte")
images10  = read_idx_file("t10k-images.idx3-ubyte")
images60  = read_idx_file("train-images.idx3-ubyte")


data10 = np.reshape(images10, (images10.shape[0], images10.shape[1]*images10.shape[2]), order='C')
data60 = np.reshape(images60, (images60.shape[0], images60.shape[1]*images60.shape[2]), order='C')
# print(data10.shape)

u, s, vt = np.linalg.svd(data60, full_matrices=False)

print(u.shape, s.shape, vt.shape)

s_inv = 1.0/s

rank = 20

sk = np.copy(s)
sk[rank:] = 0
sk_inv= np.copy(s_inv)
sk_inv[rank:] = 0

u[:,rank:] = 0

# print(data10.shape)
v = vt.transpose()

result_count = 25
query_count = 20

q_results = []
for q_id, q in enumerate(data10):
	# make representation of q in new space
	uq = np.dot(q, v) * sk_inv

	# compute similarity between query and reduces dataset
	results = []
	for idx, data in enumerate(u):
		dist = np.linalg.norm(uq-data)
		results.append((idx, dist))
	#sort results by similarity
	results = sorted(results, key=lambda tup: tup[1])
	
	q_results.append((q_id, [x[0] for x in results[:result_count]]))
	
	if (q_id==query_count): break

# print(q_results)
make_image_result(q_results, images10, images60, "results.png")

# for remain in range(1, s.shape[0]+1, 5):#[1, 5, 10, 20, 30, 50, 100, 200, s.shape[0]]
# 	sk = np.copy(s)
# 	sk[remain:] = 0
# 	recon = np.dot(u * sk, vt)
# 	recon10 = np.reshape(recon, (images10.shape[0], images10.shape[1],images10.shape[2]), order='C')
# 	recon10 = recon10.astype(np.int32)
# 	make_image_matrix(10,10, recon10, "10k_{:03d}.png".format(remain))


# print(s)
# s[20:] = 0
# print(s)



# print(recon.shape)
# print(np.allclose(data10, recon))

# print(data10[0])
# print(recon[0])

# recon10 = np.reshape(recon, (images10.shape[0], images10.shape[1],images10.shape[2]), order='C')
# recon10 = recon10.astype(np.int32)
# # print(recon10.shape)
# # print(s)
# # print(vt[0])
# # print(images10[0])

# # img = Image.fromarray(images10[0], mode='L')
# # img.save("test.png", "PNG")

# make_image_matrix(10,10, images10, "10k.png")
# make_image_matrix(10,10, recon10, "10k_r.png")
# make_image_matrix(300,200, images60, "60k.png")
