#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 23 15:36:58 2021

@author: vla04
"""
import numpy as np

A = np.matrix('2.0 -4 1; -1 -1 1; 2 -1 -2')

def scalarproduct(x,y):
    return x.transpose().dot(y)[0,0]

def norm(x):
    return  np.sqrt(scalarproduct(x,x))

def householder(A,type_):
    # returns B, Tl Tr such that A = Tl * B * Tr
    n = A.shape[0]
    Tl = np.eye(n)
    Tr = np.eye(n)
    B = A.copy()
    if (type_ == 'qr'):
        # A = Tl * B;   B is upper triangular
        for i in range(0,n-1):
            x = B[:,i].copy()
            x[0:i,0] = 0
            ei = np.zeros((n,1))
            ei[i,0] = 1
            if (scalarproduct(x,ei) > 0):
                u = x - norm(x)*ei
            else:
                u = x + norm(x)*ei
            w = (1.0/norm(u)) * u
            Hl = np.eye(n) + 2 * w * w.transpose()
            B = Hl * B
            Tl = Tl * Hl
    elif (type_ == 'bidiag'):
        # A = Tl * B * Tr;   B is bidiagonal
        for i in range(0,n-1):
            x = B[:,i].copy()  # row operations
            x[0:i,0] = 0
            ei = np.zeros((n,1))
            ei[i,0] = 1
            if (scalarproduct(x,ei) > 0):
                u = x - norm(x)*ei
            else:
                u = x + norm(x)*ei
            w = (1.0/norm(u)) * u
            Hl = np.eye(n) - 2 * w * w.transpose()
            B = Hl * B
            Tl = Tl * Hl
            if i < n-2:
                x = B[i,:].copy().transpose()  # col operations
                x[0:i+1,0] = 0
                ei = np.zeros((n,1))
                ei[i+1,0] = 1
                if (scalarproduct(x,ei) > 0):
                    u = x - norm(x)*ei
                else:
                    u = x + norm(x)*ei
                w = (1.0/norm(u)) * u
                Hr = np.eye(n) - 2 * w * w.transpose()
                B = B * Hr
                Tr = Hr * Tr
    elif (type_ == 'tridiag'):
        # A = Tl * B * Tl.transpose();   B is tridiagonal
        for i in range(0,n-2):
            x = B[:,i].copy()  # row operations
            x[0:i+1,0] = 0
            ei = np.zeros((n,1))
            ei[i+1,0] = 1
            if (scalarproduct(x,ei) > 0):
                u = x - norm(x)*ei
            else:
                u = x + norm(x)*ei
            w = (1.0/norm(u)) * u
            Hl = np.eye(n) - 2 * w * w.transpose()
            B = Hl * B * Hl
            Tl = Tl * Hl
            Tr = Tl.transpose()
    return (B,Tl,Tr)
    
#(B,Tl,Tr) = householder(A,'bidiag')
A = 0.5 * (A + A.transpose())
(B,Tl,Tr) = householder(A,'tridiag')
print(B)